from math import floor
from itertools import chain, tee
from string import whitespace
from reportlab.pdfbase.pdfmetrics import stringWidth, getFont, getAscentDescent
from .style import TextAlign, default as default_style

__all__ = ('Box', 'Block', 'Inline', 'ListItem')


def pairwise(iterable):
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def flatten(sub):
    for line in sub:
        if isinstance(line, Line):
            yield from line.words
        else:
            yield line


class StyledPart:
    __slots__ = ('style', 'part')

    def __init__(self, style, part):
        self.style = style
        self.part = part

    @property
    def width(self):
        return stringWidth(self.part, self.style.font_name, self.style.font_size)

    @property
    def height(self):
        return self.style.leading


SPACE_WIDTH_CACHE = {}


class Word:
    __slots__ = ('parts',)

    def __init__(self, style, word):
        self.parts = [StyledPart(style, word)]

    def add(self, style, part):
        if self.parts[-1].style != style:
            self.parts.append(StyledPart(style, part))
        else:
            self.parts[-1] = StyledPart(style, self.parts[-1].part+part)
        return self

    @property
    def width(self):
        return sum(p.width for p in self.parts)

    @property
    def height(self):
        return max(p.height for p in self.parts)

    @property
    def space_width(self):
        args = (self.parts[-1].style.font_name, self.parts[-1].style.font_size)
        if args not in SPACE_WIDTH_CACHE:
            SPACE_WIDTH_CACHE[args] = stringWidth(' ', *args)
        return SPACE_WIDTH_CACHE[args]

    def __str__(self):
        return ''.join(p.part for p in self.parts)


class Line:

    __slots__ = ('words', 'width', 'height')

    def __init__(self, words=None):
        self.words = words or []
        self.width = 0
        self.height = 0

    def add(self, word):
        self.words.append(word)

    @property
    def empty(self):
        return not self.words

    def __iter__(self):
        return iter(self.words)


class Behavior:
    __slots__ = ('box',)

    def __init__(self, box):
        self.box = box

    @property
    def text_allowed(self):
        return True

    def add(self, child):
        if self.box.buffered:
            self.box.children.append(child)

    def insert(self, i, child):
        if self.box.buffered:
            self.box.children.insert(i, child)

    def wrap(self, page, available_width):
        self.box.width = available_width
        self.box.height = 0

        line = None
        self.box.lines = lines = []

        max_width = 0
        for box in self.box.boxes:

            if isinstance(box, Word):

                word_width = box.width
                if not line or line.width+word_width > available_width:
                    # no line or line is full, start new line
                    line = Line()
                    lines.append(line)

                line.width += word_width + box.space_width
                line.height = max(box.height, line.height)
                max_width = max(line.width, max_width)
                line.add(box)

            elif box.tag == 'br':
                # br is added to current line and starts new line
                if not line:
                    line = Line()
                    lines.append(line)
                box.wrap(page, available_width)
                line.height = max(box.height, line.height)
                line.add(box)
                line = None

            else:
                # block elements take up their own line
                line = None
                box.wrap(page, available_width)
                lines.append(box)

        content_height = sum(b.height for b in lines)
        if self.box.tag == 'br':
            content_height = self.box.style.leading

        self.box.height = self.box.frame_height + content_height

        return max_width, self.box.height

    def split(self, top_container, bottom_container, available_height):
        lines = floor(available_height / self.box.style.leading)
        if lines == len(self.box.children):
            self.box.parent = top_container
            top_container.children.append(self)
            return self, None
        else:
            top = Box(top_container, self.box.tag, self.box.attrs)
            top.position = self.box.position
            top.position_of_type = self.box.position
            top.children = list(flatten(self.box.lines[:lines]))
            bottom = Box(bottom_container, self.box.tag, self.box.attrs)
            bottom.position = self.box.position
            bottom.position_of_type = self.box.position
            bottom.children = list(flatten(self.box.lines[lines:]))
            return top, bottom

    def draw(self, page, x, y):
        final_x, final_y = x, y - self.box.height
        style = self.box.style
        width = self.box.width
        txt = page.begin_text(final_x, final_y)
        x, y, alignment = 0, self.box.height - style.leading, style.text_align
        txt.set_position(x, y)
        txt.set_font(style.font_name, style.font_size, style.leading)
        for line in self.box.lines:
            if isinstance(line, Line):
                if style.text_align == TextAlign.right:
                    txt.set_position(width - line.width)
                elif style.text_align == TextAlign.center:
                    txt.set_position((width - line.width) / 2.0)
                fragments = []
                for word in line.words:
                    if isinstance(word, Word):
                        for part in word.parts:
                            if style != part.style:
                                if fragments:
                                    txt.draw(''.join(fragments))
                                    fragments = []
                                style = part.style
                                txt.set_font(style.font_name, style.font_size, style.leading)
                            fragments.append(part.part)
                        fragments.append(' ')
                if fragments and fragments[-1] == ' ':
                    fragments.pop()  # last space
                txt.draw(''.join(fragments), new_line=True)
            else:
                # non-text line (probably some block element, list item or table)
                # TODO: calculate y offset after some text lines have been written
                #       currently vertical movement is handled internally by PDF 'T*' command
                _, y = line.draw(page, x, y)
        txt.close()
        return final_x, final_y


class Block(Behavior):
    __slots__ = ()


class Inline(Behavior):
    __slots__ = ()


class ListItem(Behavior):
    __slots__ = ()


class Box:

    __slots__ = (
        'parent', 'tag', 'attrs', 'id', 'classes', 'style', 'behavior', 'buffered',
        'position', 'last', 'position_of_type', 'last_of_type', 'width', 'height',
        'children', 'lines', 'before', 'after'
    )

    def __init__(self, parent, tag, attrs):
        self.parent = parent
        self.tag = tag
        self.attrs = attrs
        self.id = attrs.get('id', None)
        self.classes = attrs.get('class', [])

        if parent is None:
            self.style = default_style
        else:
            self.style = parent.style.inherit()

        self.behavior = None

        self.position = 1
        self.position_of_type = 1
        self.last = False

        self.width = None
        self.height = None

        self.children = []
        self.lines = []

        self.before = None
        self.after = None

        self.buffered = True

    def __str__(self):
        return '{}: <{}>'.format(self.behavior.__class__.__name__, self.tag)

    @property
    def first(self):
        return self.position == 1

    @property
    def first_of_type(self):
        return self.position_of_type == 1

    @property
    def text_allowed(self):
        return self.behavior.text_allowed

    def add(self, child):
        self.behavior.add(child)

    def insert(self, i, child):
        self.behavior.insert(i, child)

    def draw_border_and_background(self, page, x, y):
        page.save_state()
        page.translate(x, y)
        s = self.style
        top_left, top_right, bottom_left, bottom_right = self.border_box
        if s.background_color:
            page.fill_color(*s.background_color)
            page.rectangle(0, 0, self.width, self.height)
        if s.border_top_width > 0:
            page.line_width(s.border_top_width)
            page.stroke_color(*s.border_top_color)
            page.line(*top_left, *top_right)
        if s.border_right_width > 0:
            page.line_width(s.border_right_width)
            page.stroke_color(*s.border_right_color)
            page.line(*top_right, *bottom_right)
        if s.border_bottom_width > 0:
            page.line_width(s.border_bottom_width)
            page.stroke_color(*s.border_bottom_color)
            page.line(*bottom_left, *bottom_right)
        if s.border_left_width > 0:
            page.line_width(s.border_left_width)
            page.stroke_color(*s.border_left_color)
            page.line(*top_left, *bottom_left)
        page.restore_state()

    @property
    def empty(self):
        return not self.children

    @property
    def min_content_width(self):
        width = 0
        for block in self.children:
            width = max(width, block.min_content_width)
        return width

    @property
    def frame_top(self):
        s = self.style
        return s.margin_top + s.border_top_width + s.padding_top

    @property
    def frame_right(self):
        s = self.style
        return s.padding_right + s.border_right_width + s.margin_right

    @property
    def frame_bottom(self):
        s = self.style
        return s.margin_bottom + s.border_bottom_width + s.padding_bottom

    @property
    def frame_left(self):
        s = self.style
        return s.margin_left + s.border_left_width + s.padding_left

    @property
    def frame_width(self):
        return self.frame_left + self.frame_right

    @property
    def frame_height(self):
        return self.frame_top + self.frame_bottom

    @property
    def border_box(self):
        s = self.style
        return (
            (s.margin_left, self.height-s.margin_top),  # top left
            (self.width-s.margin_right, self.height-s.margin_top),  # top right
            (s.margin_left, s.margin_bottom),  # bottom left
            (self.width-s.margin_right, s.margin_bottom),  # bottom right
        )

    @property
    def min_content_width(self):
        return 0

    def traverse(self):
        for child in self.children:
            if isinstance(child, str):
                yield child, self
            elif child.style.display == 'inline' and not child.tag == 'br':
                yield from child.traverse()
            else:
                yield child, self

    @property
    def boxes(self):
        words = []
        word_open = False
        for box, parent in self.traverse():
            if isinstance(box, str):
                if box[0] in whitespace:
                    word_open = False
                parts = box.split()
                if not parts:
                    continue
                word_iter = iter(parts)
                if word_open:
                    words[-1].add(parent.style, next(word_iter))
                words.extend(Word(parent.style, word) for word in word_iter)
                word_open = box[-1] not in whitespace
            else:
                if words:
                    yield from words
                    words = []
                    word_open = False
                yield box
        if words:
            yield from words

    def wrap(self, page, available_width):
        return self.behavior.wrap(page, available_width)

    def split(self, top_container, bottom_container, available_height):
        return self.behavior.split(top_container, bottom_container, available_height)

    def draw(self, page, x, y):
        return self.behavior.draw(page, x, y)



from enum import Enum
from collections import namedtuple

from reportlab.lib.fonts import tt2ps
from reportlab.lib.colors import Color, black


__all__ = (
    'Style', 'TextAlign', 'VerticalAlign',
    'BorderStyle', 'BorderCollapse', 'default'
)


class TextAlign:
    left = 'left'
    center = 'center'
    right = 'right'
    justify = 'justify'


class VerticalAlign:
    top = 'top'
    middle = 'middle'
    bottom = 'bottom'


class BorderStyle(Enum):
    solid = 1
    dashed = 2


class BorderCollapse(Enum):
    collapse = 1
    separate = 2


class Style(namedtuple('_Style', (

        'display', 'visibility',

        'font_family', 'font_size', 'font_weight', 'font_style', 'text_decoration', 'color',

        'text_align', 'vertical_align',

        'padding_top', 'padding_right', 'padding_bottom', 'padding_left',

        'border_top_color', 'border_top_style', 'border_top_width',
        'border_left_color', 'border_left_style', 'border_left_width',
        'border_right_color', 'border_right_style', 'border_right_width',
        'border_bottom_color', 'border_bottom_style', 'border_bottom_width',

        'margin_top', 'margin_right', 'margin_bottom', 'margin_left',

        'background_color',

        'page_break_before',
        'page_break_inside',
        'page_break_after',

        'page_bottom_right_content',

        'border_collapse',
        'border_spacing_horizontal',
        'border_spacing_vertical',

        'letterhead_page'

        ))):

    def set(self, **attrs):
        return self._replace(**self.clean(attrs))

    def inherit(self):
        return default.set(
            font_family=self.font_family,
            font_size=self.font_size,
            font_weight=self.font_weight,
            font_style=self.font_style,
            text_decoration=self.text_decoration,
            text_align=self.text_align,
            color=self.color,
            # TODO: add all inheritable attributes
        )

    @staticmethod
    def clean(kwargs):
        if 'page_break_before' in kwargs:
            kwargs['page_break_before'] = kwargs['page_break_before'] == 'always'
        if 'page_break_after' in kwargs:
            kwargs['page_break_after'] = kwargs['page_break_after'] == 'always'
        return kwargs

    @property
    def bold(self):
        return 1 if self.font_weight == 'bold' else 0

    @property
    def italic(self):
        return 1 if self.font_style == 'italic' else 0

    @property
    def font_name(self):
        return tt2ps(self.font_family, self.bold, self.italic)

    @property
    def leading(self):
        return int(self.font_size * 1.2)


default = Style(

    display=None,
    visibility=True,

    font_family='Helvetica',
    font_size=12,
    font_weight='',
    font_style='',
    text_decoration='',
    color=(0, 0, 0, 1),

    text_align='left',

    vertical_align='top',

    padding_top=0,
    padding_right=0,
    padding_bottom=0,
    padding_left=0,

    border_top_color=(0, 0, 0, 1),
    border_top_style=BorderStyle.solid,
    border_top_width=0,
    border_right_color=(0, 0, 0, 1),
    border_right_style=BorderStyle.solid,
    border_right_width=0,
    border_bottom_color=(0, 0, 0, 1),
    border_bottom_style=BorderStyle.solid,
    border_bottom_width=0,
    border_left_color=(0, 0, 0, 1),
    border_left_style=BorderStyle.solid,
    border_left_width=0,

    margin_top=0,
    margin_right=0,
    margin_bottom=0,
    margin_left=0,

    background_color=None,

    page_break_before=False,
    page_break_inside=True,
    page_break_after=False,

    page_bottom_right_content='',

    border_collapse=BorderCollapse.separate,
    border_spacing_horizontal=2,
    border_spacing_vertical=2,

    letterhead_page=1

    # TODO: add all CSS attributes
)



from .style import BorderCollapse
from .box import Box, Behavior


__all__ = (
    'Table',
    'TableColumnGroup', 'TableColumn',
    'TableRowGroup', 'TableRow', 'TableCell'
)


TABLE_CHILDREN = {
    'table-column-group': 0,
    'table-header-group': 1,
    'table-row-group': 2,
    'table-footer-group': 3
}


class Table(Behavior):
    __slots__ = ()

    def __init__(self, box):
        super().__init__(box)
        columns = Box(self.box, 'colgroup', {})
        box.children = [columns, None, None, None]
        columns.style = columns.style.set(display='table-column-group')
        columns.behavior = TableColumnGroup(columns)

    @property
    def text_allowed(self):
        return False

    @property
    def columns(self):
        return self.box.children[TABLE_CHILDREN['table-column-group']]

    @property
    def thead(self):
        return self.box.children[TABLE_CHILDREN['table-header-group']]

    @property
    def tbody(self):
        return self.box.children[TABLE_CHILDREN['table-row-group']]

    @property
    def tfoot(self):
        return self.box.children[TABLE_CHILDREN['table-footer-group']]

    def add(self, child):
        i = TABLE_CHILDREN[child.style.display]
        if i != TABLE_CHILDREN['table-column-group']:
            assert self.box.children[i] is None
        if i == TABLE_CHILDREN['table-row-group']:
            child.buffered = self.box.parent.tag != 'body'
        else:
            child.style = child.style.set(visibility=False)
            if i == TABLE_CHILDREN['table-column-group']:
                child.behavior.measurements = self.columns.behavior.measurements
        self.box.children[i] = child

    def get_column_widths(self, available_width):
        return self.columns.behavior.get_widths(available_width)

    def reserve_header_height(self, page, available_width):
        if self.thead and self.thead.behavior.drawn_on != page.page_number:
            return self.thead.wrap(page, available_width)
        return 0

    def reserve_footer_height(self, page, available_width):
        if self.tfoot:
            return self.tfoot.wrap(page, available_width)
        return 0

    def wrap(self, page, available_width):
        self.height = (
            self.reserve_header_height(page, available_width) +
            self.tbody.wrap(page, available_width) +
            self.reserve_header_height(page, available_width)
        )
        self.width = available_width
        return self.width, self.height

    def draw_header(self, page, x, y):
        if self.thead and self.thead.behavior.drawn_on != page.page_number:
            return self.thead.draw(page, x, y)
        return x, y

    def draw_footer(self, page, x, y):
        if self.tfoot:
            return self.tfoot.draw(page, x, y)
        return x, y

    def draw(self, page, x, y):
        x, y = self.draw_header(page, x, y)
        x, y = self.tbody.draw(page, x, y)
        x, y = self.draw_footer(page, x, y)
        return x, y

    def table_old_wrap(self, available_width, _=None):
        collapsed = self.style.border_collapse == BorderCollapse.collapse
        columns_available_width = available_width - self.frame_width
        vspacing = hspacing = 0
        if not collapsed:
            vspacing = self.style.border_spacing_vertical
            hspacing = self.style.border_spacing_horizontal
            columns_available_width -= hspacing*len(self.columns)
        self.calculate_ratio_columns(columns_available_width)
        self.width = available_width
        self.height = self.frame_height
        self.content_heights = []
        for row in self.rows:
            row.collapsed = collapsed
            row.cell_spacing = hspacing
            _, height = row.wrap(self.content_widths)
            height += vspacing
            self.content_heights.append(height)
            self.height += height
        return available_width, self.height

    def table_old_split(self, available_width, available_height):
        if self.content_heights is None:
            self.wrap(available_width)

        assert self.content_heights, "Split called on empty table."

        if self.tfoot:

            # check that footer fits, exit fast if not
            footer_height = sum(self.content_heights[-len(self.tfoot):])
            if footer_height >= available_height:
                return []

            # reduce the available height by footer to make sure there is always room for it
            available_height -= footer_height

        heights = iter(self.content_heights)

        # now check if header also fits, exit fast if not
        header_height = 0
        for _, height in zip(self.thead, heights):
            header_height += height
            if header_height >= available_height:
                return []

        # finally lets see if some rows fit sandwiched between thead/tfoot
        split_at_row = -1
        split_row_height = 0
        consumed_height = header_height
        for split_at_row, (row, split_row_height) in enumerate(zip(self.content, heights)):
            consumed_height += split_row_height
            if consumed_height >= available_height:
                break

        if consumed_height <= available_height:
            if len(self.content) == 1 or len(self.content)-1 == split_at_row:
                # nothing to split
                return [self]
            else:
                # table split cleanly between rows, no need to further split any rows/cells
                return [
                    Table(self.ratios, self.columns, self.content[:split_at_row+1], self.style, self.thead, self.tfoot, was_split=True),
                    Table(self.ratios, self.columns, self.content[split_at_row-1:], self.style, self.thead, self.tfoot, was_split=True)
                ]

        top_rows = self.content[:split_at_row]
        row = self.content[split_at_row]
        bottom_rows = self.content[split_at_row+1:]

        top_height = available_height - (consumed_height - split_row_height)
        if self.style.border_collapse == BorderCollapse.separate:
            top_height -= self.style.border_spacing_vertical
        split = row.splitOn(None, available_width, top_height)

        if not split:
            # cell could not be split,
            # move entirely to bottom half
            bottom_rows.insert(0, row)
        elif len(split) == 1:
            # cell completely fits in top half,
            # add placeholder to bottom half
            top_rows.append(row)
        elif len(split) == 2:
            top_rows.append(split[0])
            bottom_rows.insert(0, split[1])
        else:
            raise LayoutError("Splitting row {} produced unexpected result.".format(split_at_row))

        return [] if not top_rows else [
            Table(self.ratios, self.columns, top_rows, self.style, self.thead, self.tfoot, was_split=True),
            Table(self.ratios, self.columns, bottom_rows, self.style, self.thead, self.tfoot, was_split=True)
        ]


class TableColumnGroup(Behavior):
    __slots__ = ('widths', 'measurements', 'span')

    def __init__(self, box):
        super().__init__(box)
        self.span = int(box.attrs.get('span', 1))
        self.widths = None
        self.measurements = []

    @property
    def text_allowed(self):
        return False

    @property
    def count(self):
        if self.span is not None:
            return self.span
        return len(self.box.children)

    def get_widths(self, available_width):
        if self.widths is None:
            if self.measurements:
                self.calculate_widths(available_width)
            else:
                self.widths = [available_width/self.span] * self.span
        return self.widths

    def calculate_widths(self, available_width):

        widths = self.measurements.copy()

        fixed = sum(widths)

        units = 0
        for c in self.box.children:
            if c.behavior.is_proportional:
                units += c.behavior.proportion

        unit_size = (available_width - fixed) / units
        for i in range(len(self.box.children)):
            c = self.box.children[i]
            if c.behavior.is_proportional:
                widths[i] = c.behavior.proportion * unit_size

        self.widths = widths

    def measure(self, row, data):
        maximums = data['maximums']
        if not self.box.children:
            data['span'] = max(data['span'], len(row.children))
            return
        cols = iter(self.box.children)
        icol = 0
        for cell in row.children:
            if not cell or cell.behavior.colspan == 1:
                col = next(cols)
                if cell and cell.children:
                    width = col.behavior.measure(cell) + cell.frame_width
                    maximums[icol] = max(maximums[icol], width)
                icol += 1
            else:
                # multi-column spans aren't measured
                for _ in range(cell.behavior.colspan):
                    next(cols)
                    icol += 1


class TableColumn(Behavior):
    __slots__ = ('width',)

    def __init__(self, box):
        super().__init__(box)
        self.width = box.attrs.get('width', '')

    @property
    def text_allowed(self):
        return False

    @property
    def is_fixed(self):
        return all(n in '0123456789' for n in self.width)

    @property
    def fixed(self):
        return int(self.width)

    @property
    def is_max_content(self):
        return self.width in ('0*', 'max-content')

    @property
    def is_min_content(self):
        return self.width in ('min-content',)

    @property
    def is_proportional(self):
        return self.width.endswith('*') and self.width != '0*'

    @property
    def proportion(self):
        return int(self.width[:-1])

    def measure(self, row):
        if self.is_fixed:
            return self.fixed
        elif self.is_max_content:
            return row.wrap(None, 1000)[0]
        elif self.is_min_content:
            return row.wrap(None, 0)[0]
        return 0


class TableRowGroup(Behavior):
    __slots__ = ('drawn_on',)

    def __init__(self, box):
        super().__init__(box)
        self.drawn_on = None

    @property
    def text_allowed(self):
        return False

    def add(self, child):
        if self.box.parent.buffered:
            self.box.children.append(child)

    def wrap(self, page, available_width):
        header_height = 0
        for row in self.box.children:
            header_height += row.behavior._wrap(page, available_width)[1]
        return header_height

    def draw(self, page, x, y):
        self.drawn_on = page.page_number
        for row in self.box.children:
            x, y = row.behavior._draw(page, x, y)
        self.box.position += 1
        return x, y


class TableRow(Behavior):
    __slots__ = ()

    @property
    def text_allowed(self):
        return False

    @property
    def table_row_frame_top(self):
        if self.collapsed:
            return self.style.border_top_width
        return 0

    @property
    def table_row_frame_right(self):
        return 0

    @property
    def table_row_frame_bottom(self):
        if self.collapsed:
            return self.style.border_bottom_width
        return 0

    @property
    def table_row_frame_left(self):
        return 0

    def wrap(self, page, available_width):
        table = self.box.parent.parent.behavior
        return available_width, (
            table.reserve_header_height(page, available_width) +
            self._wrap(page, available_width)[1] +
            table.reserve_footer_height(page, available_width)
        )

    def _wrap(self, page, available_width):
        table = self.box.parent.parent.behavior
        children = self.box.children
        column_widths = table.get_column_widths(available_width)
        self.box.width = available_width
        cell_widths = self.box.lines = []

        if not children:
            return 0, 0

        col_width = iter(column_widths)
        for cell in children:
            if not cell or cell.behavior.colspan == 1:
                cell_widths.append(next(col_width))
            else:
                cell_widths.append(0)
                for col in range(cell.behavior.colspan):
                    # TODO: re-implement cell_spacing
                    cell_widths[-1] += next(col_width)  # + self.box.cell_spacing / 2.0

        max_height = 0
        for cell, cell_width in zip(children, cell_widths):
            if cell is not None:
                # TODO: re-implement collapsed
                #cell.collapsed = self.collapsed
                _, height = cell.wrap(page, cell_width)
                max_height = max(max_height, height)

        for cell in children:
            if not cell: continue
            cell.height = max_height

        self.box.height = self.box.frame_height + max_height
        return self.box.width, self.box.height

    def split(self, top_parent, bottom_parent, available_height):
        if available_height >= self.box.height:
            return self, None

        content_height = available_height - self.box.frame_height
        top_half = Box(self.box.parent, self.box.tag, self.box.attrs)
        top_half.behavior = self.box.behavior
        top_half.position = self.box.position
        top_half.position_of_type = self.box.position_of_type
        bottom_half = Box(self.box.parent, self.box.tag, self.box.attrs)
        bottom_half.behavior = self.box.behavior
        bottom_half.position = self.box.position+1
        bottom_half.position_of_type = self.box.position_of_type+1

        for cell in self.box.children:
            if cell:
                cell.split(top_half, bottom_half, content_height)

        if any(c.children for c in bottom_half.children if c):
            bottom_half.last = self.box.last
            return top_half, bottom_half
        top_half.last = self.box.last
        return top_half, None

    def draw(self, page, x, y):
        table = self.box.parent.parent
        x, y = table.behavior.draw_header(page, x, y)
        x, y = self._draw(page, x, y)
        return x, y

    def _draw(self, page, x, y):
        original_x, original_y = x, y
        # TODO: re-implement cell_spacing
        #x += self.cell_spacing / 2.0
        #y -= self.height
        y -= self.box.frame_top
        children = self.box.children
        cell_widths = self.box.lines
        last = len(cell_widths)-1
        for i, (cell, width) in enumerate(zip(children, cell_widths)):
            if cell:
                #if False and self.collapsed:
                #    self.draw_collapsed_cell_border(
                #        None if i == 0 else self.children[i-1],
                #        cell,
                #        None if i == last else self.children[i+1],
                #    )
                #else:
                cell.draw_border_and_background(page, x, original_y - self.box.height)
                cell.draw(page, x, y)
            # TODO: re-implement cell_spacing
            x += width #+ self.cell_spacing / 2.0
        return original_x, original_y - self.box.height

    def draw_collapsed_cell_border(self, before, cell, after):
        pass


class TableCell(Behavior):
    __slots__ = ('colspan',)

    def __init__(self, box):
        super().__init__(box)
        self.colspan = box.attrs.get('colspan', None)
        if self.colspan is not None:
            self.colspan = int(self.colspan)
            assert self.colspan > 0
        else:
            self.colspan = 1

    @property
    def cell_frame_top(self):
        s = self.style
        if self.collapsed:
            return s.border_top_width/2 + s.padding_top
        else:
            return s.border_top_width + s.padding_top

    @property
    def cell_frame_right(self):
        s = self.style
        if self.collapsed:
            return s.padding_right + s.border_right_width/2.0
        else:
            return s.padding_right + s.border_right_width

    @property
    def cell_frame_bottom(self):
        s = self.style
        if self.collapsed:
            return s.border_bottom_width/2.0 + s.padding_bottom
        else:
            return s.border_bottom_width + s.padding_bottom

    @property
    def cell_frame_left(self):
        s = self.style
        if self.collapsed:
            return s.border_left_width/2.0 + s.padding_left
        else:
            return s.border_left_width + s.padding_left

    @property
    def cell_border_box(self):
        return (
            (0, self.height),  # top left
            (self.width, self.height),  # top right
            (0, 0),  # bottom left
            (self.width, 0),  # bottom right
        )

    def cell_draw(self, page, x, y):
        x += self.frame_left
        if self.style.vertical_align == VerticalAlign.top:
            y -= self.frame_top
        elif self.style.vertical_align == VerticalAlign.middle:
            y -= (self.height + sum(self.content_heights)) / 2.0
        else:
            y -= self.height - (sum(self.content_heights) + self.frame_bottom)

        for block, height in zip(self.children, self.content_heights):
            block.draw(page, x, y)
            y -= height

    def cell_wrap(self, page, available_width):
        self.width = available_width
        self.content_heights = []

        self.css.apply(self)
        if self.parent:
            self.style = self.style.inherit(self.parent.style)

        content_width = available_width - self.frame_width

        consumed = 0
        for block in self.children:
            _, height = block.wrap(page, content_width)
            consumed += height
            self.content_heights.append(height)

        self.height = self.frame_height + consumed
        return available_width, self.height

    def cell_split(self, top_parent, bottom_parent, available_height):

        if not self.content_heights:
            self.parent = top_parent
            top_parent.children.append(self)
            bottom_parent.children.append(None)
            return

        if self.style.vertical_align == VerticalAlign.top:
            consumed_height = self.frame_top
        elif self.style.vertical_align == VerticalAlign.middle:
            consumed_height = (self.height - sum(self.content_heights)) / 2.0
        else:
            consumed_height = self.height - (sum(self.content_heights) + self.frame_bottom)

        if consumed_height >= available_height:
            # border and padding don't even fit
            top_parent.children.append(None)
            self.parent = bottom_parent
            bottom_parent.children.append(self)
            return

        split_at_block, split_block_height = -1, 0
        for split_at_block, split_block_height in enumerate(self.content_heights):
            consumed_height += split_block_height
            if consumed_height >= available_height:
                break

        if consumed_height <= available_height:
            if len(self.children) == 1 or len(self.children)-1 == split_at_block:
                self.parent = top_parent
                top_parent.children.append(self)
                bottom_parent.children.append(None)
                return
            else:
                top_cell = Cell(self.tag, top_parent, self.id, self.classes, self.css, self.position, colspan=self.colspan)
                for top_child in self.children[:split_at_block+1]:
                    top_child.parent = top_cell
                    top_cell.children.append(top_child)
                bottom_cell = Cell(self.tag, bottom_parent, self.id, self.classes, self.css, self.position, colspan=self.colspan)
                for bottom_child in self.children[split_at_block-1:]:
                    bottom_child.parent = bottom_cell
                    bottom_cell.children.append(bottom_child)
                return top_cell, bottom_cell

        top_cell = Cell(self.tag, top_parent, self.id, self.classes, self.css, self.position, colspan=self.colspan)
        for top_child in self.children[:split_at_block]:
            top_child.parent = top_cell
            top_cell.children.append(top_child)

        block = self.children[split_at_block]
        bottom_cell = Cell(self.tag, bottom_parent, self.id, self.classes, self.css, self.position, colspan=self.colspan)

        block.split(top_cell, bottom_cell, available_height - (consumed_height - split_block_height))

        for bottom_child in self.children[split_at_block+1:]:
            bottom_child.parent = bottom_cell
            bottom_cell.children.append(bottom_child)

        return top_cell, bottom_cell



from lxml import etree
from .box import *
from .table import *

__all__ = ('HTMLParser',)


TAG_TO_DISPLAY = {
    'table': 'table',
    'colgroup': 'table-column-group',
    'col': 'table-column',
    'thead': 'table-header-group',
    'tfoot': 'table-footer-group',
    'tbody': 'table-row-group',
    'tr': 'table-row',
    'td': 'table-cell',
    'th': 'table-cell',

    'body': 'block',
    'div': 'block',
    'p': 'block',
    'ul': 'block',
    'ol': 'block',

    'li': 'list-item',

    'span': 'inline',
    'a': 'inline',
    'b': 'inline',
    'strong': 'inline',
    'i': 'inline',
    'u': 'inline',
    'br': 'inline',
}


def pumper(html_generator):
    """
    Pulls HTML from source generator,
    feeds it to the parser and yields
    DOM elements.
    """
    source = html_generator()
    parser = etree.HTMLPullParser(
        events=('start', 'end'),
        remove_comments=True
    )
    while True:
        for element in parser.read_events():
            yield element
        try:
            parser.feed(next(source))
        except StopIteration:
            # forces close of any unclosed tags
            parser.feed('</html>')
            for element in parser.read_events():
                yield element
            break


def iter_previous_current_last(child_iter):
    current_child = next(child_iter, None)
    if current_child:
        previous_child = None
        for next_child in child_iter:
            yield previous_child, current_child, False
            previous_child = current_child
            current_child = next_child
        yield previous_child, current_child, True


class GrowingList(list):
    def __getitem__(self, index):
        if index >= len(self):
            self.extend([0]*(index + 1 - len(self)))
        return super().__getitem__(index)

    def __setitem__(self, index, value):
        if index >= len(self):
            self.extend([0]*(index + 1 - len(self)))
        super().__setitem__(index, value)


class HTMLParser:

    def __init__(self, html_generator, css=None):
        self.html_generator = html_generator
        self.css = css
        self.table_columns = None
        self.stream = None

    def parse(self, parent):
        """
        https://www.w3.org/TR/html5/syntax.html
        See: 8.2.5 Tree construction
        """

        last_child = None

        for event, element in self.stream:

            tag = element.tag

            if event == 'start':

                last_child = element

                if isinstance(parent.behavior, Table):

                    # these are not direct descendants of <table>, add missing parent

                    if tag in ('td', 'th', 'tr'):
                        parent = parent.behavior.tbody or self.make_child(parent, 'tbody', {})

                    elif tag == 'col':
                        parent = parent.behavior.columns or self.make_child(parent, 'colgroup', {})

                elif isinstance(parent.behavior, TableRowGroup):

                    # these are not direct descendants of <tbody>, add missing parent

                    if tag in ('td', 'th'):
                        parent = self.make_child(parent, 'tr', {})

                node = self.make_child(parent, tag, element.attrib)

                if isinstance(node.behavior, Table) and self.table_columns:
                    measurements_data = self.table_columns.pop(0)
                    node.behavior.columns.behavior.measurements = measurements_data['maximums']
                    node.behavior.columns.behavior.span = measurements_data['span']

                if isinstance(node.behavior, Table) and parent.tag == 'body':
                    yield from self.parse(node)

                else:

                    # root tables are a very special case
                    # where the table and row groups
                    # are not yielded but the table rows
                    # themselves are yielded as "root nodes"

                    skip_root_table_meta_groups = (  # don't yield <colgroup>, <thead>, <tbody>, <tfoot>
                        parent.parent and parent.parent.tag == 'body' and
                        isinstance(node.behavior, (TableRowGroup, TableColumnGroup))
                    )

                    yield_root_table_rows = (  # do yield <tr> that's inside <tbody>
                        skip_root_table_meta_groups and
                        node.style.display == 'table-row-group'
                    )

                    position = 0
                    position_of_type = {}

                    for previous, (child_element, child_node), last in iter_previous_current_last(self.parse(node)):
                        if previous is not None:
                            if node.text_allowed and previous[0].tail:
                                # insert text between previous node and current node
                                node.insert(len(node.children)-2, previous[0].tail)
                        position_of_type.setdefault(child_node.tag, 0)
                        child_node.position = position = position + 1
                        child_node.position_of_type = position_of_type[child_node.tag] = position_of_type[child_node.tag] + 1
                        child_node.last = last
                        if yield_root_table_rows and isinstance(child_node.behavior, TableRow):
                            yield child_element, child_node

                    if node.text_allowed and element.text:
                        node.insert(0, element.text)

                    if not skip_root_table_meta_groups:
                        yield element, node

            else:

                if parent.text_allowed and last_child is not None and last_child.tail:
                    parent.add(last_child.tail)

                break

    def make_child(self, parent, tag, attrib):
        node = Box(parent, tag, attrib)
        self.css.apply(node)
        if node.style.display is None:
            display = TAG_TO_DISPLAY.get(tag, 'block')
            node.style = node.style.set(display=display)
        node.behavior = self.get_behavior(parent.behavior if parent else None, node)
        if parent and parent.buffered:
            parent.add(node)
        return node

    @staticmethod
    def get_behavior(parent, node):

        display = node.style.display

        if display == 'block':
            return Block(node)

        elif display == 'inline':
            return Inline(node)

        elif display == 'table':
            return Table(node)

        elif parent:

            if isinstance(parent, Table):
                if display == 'table-column-group':
                    return TableColumnGroup(node)

                elif display in ('table-header-group', 'table-row-group', 'table-footer-group'):
                    return TableRowGroup(node)

            elif display == 'table-column' and isinstance(parent, TableColumnGroup):
                return TableColumn(node)

            elif display == 'table-row' and isinstance(parent, TableRowGroup):
                return TableRow(node)

            elif display == 'table-cell' and isinstance(parent, TableRow):
                return TableCell(node)

            elif display == 'list-item' and isinstance(parent, Block):
                return ListItem(node)

        return Block(node)

    def measure_column_widths(self):
        table_columns = []
        columns = None
        for box in self.boxes():
            if isinstance(box.behavior, TableRow):
                table = box.parent.parent.behavior
                if table.columns.behavior is not columns:
                    columns = table.columns.behavior
                    table_columns.append({
                        'maximums': GrowingList(),
                        'span': 1
                    })
                    if table.thead:
                        for hrow in table.thead.children:
                            columns.measure(hrow, table_columns[-1])
                    if table.tfoot:
                        for frow in table.tfoot.children:
                            columns.measure(frow, table_columns[-1])
                columns.measure(box, table_columns[-1])
        return table_columns

    def boxes(self):
        self.stream = pumper(self.html_generator)
        next(self.stream)  # html
        _, body_element = next(self.stream)  # body
        node = self.make_child(None, body_element.tag, body_element.attrib)
        node.buffered = False
        return map(lambda r: r[1], self.parse(node))

    def __iter__(self):
        if self.table_columns is None:
            self.table_columns = self.measure_column_widths()
        yield from self.boxes()
